# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: continual/rpc/logging/v1/logging.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from continual.rpc.management.v1 import management_types_pb2 as continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='continual/rpc/logging/v1/logging.proto',
  package='continual.rpc.logging.v1',
  syntax='proto3',
  serialized_options=b'ZCgithub.com/continualiq/continual/continual/rpc/logging/v1;loggingv1',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n&continual/rpc/logging/v1/logging.proto\x12\x18\x63ontinual.rpc.logging.v1\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1cgoogle/api/annotations.proto\x1a\x1bgoogle/protobuf/empty.proto\x1a\x32\x63ontinual/rpc/management/v1/management_types.proto\"\'\n\x13LogHeartbeatRequest\x12\x10\n\x08resource\x18\x01 \x01(\t\"l\n\x15LogExperimentsRequest\x12\x15\n\rmodel_version\x18\x01 \x01(\t\x12<\n\x0b\x65xperiments\x18\x02 \x03(\x0b\x32\'.continual.rpc.management.v1.Experiment\"i\n\x11LogMetricsRequest\x12\x15\n\rmodel_version\x18\x01 \x01(\t\x12=\n\x07metrics\x18\x02 \x03(\x0b\x32,.continual.rpc.management.v1.TrainingMetrics\"\x9f\x01\n LogEvaluationModelMetricsRequest\x12\x15\n\rmodel_version\x18\x01 \x01(\t\x12\x64\n evaluation_model_version_metrics\x18\x17 \x01(\x0b\x32:.continual.rpc.management.v1.EvaluationModelVersionMetrics\"l\n\x16LogDatasetStatsRequest\x12\x10\n\x08resource\x18\x01 \x01(\t\x12@\n\rdataset_stats\x18\x02 \x01(\x0b\x32).continual.rpc.management.v1.DatasetStats\"z\n\x19LogPredictionStatsRequest\x12\x18\n\x10\x62\x61tch_prediction\x18\x01 \x01(\t\x12\x43\n\x10prediction_stats\x18\x02 \x01(\x0b\x32).continual.rpc.management.v1.DatasetStats\"\x88\x01\n GenerateArtifactUploadURLRequest\x12\x10\n\x08resource\x18\x01 \x01(\t\x12\x0c\n\x04path\x18\x02 \x01(\t\x12\x0c\n\x04type\x18\x03 \x01(\t\x12\x11\n\tmime_type\x18\x04 \x01(\t\x12\x10\n\x08metadata\x18\x05 \x01(\t\x12\x11\n\tresumable\x18\x06 \x01(\x08\"i\n!GenerateArtifactUploadURLResponse\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x37\n\x08\x61rtifact\x18\x02 \x01(\x0b\x32%.continual.rpc.management.v1.Artifact\"f\n\x14LogDataChecksRequest\x12\x10\n\x08resource\x18\x01 \x01(\t\x12<\n\x0b\x64\x61ta_checks\x18\x02 \x01(\x0b\x32\'.continual.rpc.management.v1.DataChecks\"p\n\x16LogModelVersionRequest\x12@\n\rmodel_version\x18\x01 \x01(\x0b\x32).continual.rpc.management.v1.ModelVersion\x12\x14\n\x0cupdate_paths\x18\x02 \x03(\t\"y\n\x19LogBatchPredictionRequest\x12\x46\n\x10\x62\x61tch_prediction\x18\x01 \x01(\x0b\x32,.continual.rpc.management.v1.BatchPrediction\x12\x14\n\x0cupdate_paths\x18\x02 \x03(\t\"\"\n\x12GetArtifactRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"9\n\x18\x42\x61tchGetArtifactsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\r\n\x05names\x18\x02 \x03(\t\"U\n\x19\x42\x61tchGetArtifactsResponse\x12\x38\n\tartifacts\x18\x01 \x03(\x0b\x32%.continual.rpc.management.v1.Artifact\"%\n\x15\x44\x65leteArtifactRequest\x12\x0c\n\x04name\x18\x01 \x01(\t2\xa3\x0c\n\nLoggingAPI\x12l\n\x0cLogHeartbeat\x12-.continual.rpc.logging.v1.LogHeartbeatRequest\x1a\x16.google.protobuf.Empty\"\x15\x82\xd3\xe4\x93\x02\x0f\"\r/v1/heartbeat\x12r\n\x0eLogExperiments\x12/.continual.rpc.logging.v1.LogExperimentsRequest\x1a\x16.google.protobuf.Empty\"\x17\x82\xd3\xe4\x93\x02\x11\"\x0f/v1/experiments\x12\x66\n\nLogMetrics\x12+.continual.rpc.logging.v1.LogMetricsRequest\x1a\x16.google.protobuf.Empty\"\x13\x82\xd3\xe4\x93\x02\r\"\x0b/v1/metrics\x12\x8e\x01\n\x19LogEvaluationModelMetrics\x12:.continual.rpc.logging.v1.LogEvaluationModelMetricsRequest\x1a\x16.google.protobuf.Empty\"\x1d\x82\xd3\xe4\x93\x02\x17\"\x15/v1/evaluationMetrics\x12u\n\x0fLogDatasetStats\x12\x30.continual.rpc.logging.v1.LogDatasetStatsRequest\x1a\x16.google.protobuf.Empty\"\x18\x82\xd3\xe4\x93\x02\x12\"\x10/v1/datasetStats\x12~\n\x12LogPredictionStats\x12\x33.continual.rpc.logging.v1.LogPredictionStatsRequest\x1a\x16.google.protobuf.Empty\"\x1b\x82\xd3\xe4\x93\x02\x15\"\x13/v1/predictionStats\x12\x62\n\x0bGetArtifact\x12,.continual.rpc.logging.v1.GetArtifactRequest\x1a%.continual.rpc.management.v1.Artifact\x12|\n\x11\x42\x61tchGetArtifacts\x12\x32.continual.rpc.logging.v1.BatchGetArtifactsRequest\x1a\x33.continual.rpc.logging.v1.BatchGetArtifactsResponse\x12Y\n\x0e\x44\x65leteArtifact\x12/.continual.rpc.logging.v1.DeleteArtifactRequest\x1a\x16.google.protobuf.Empty\x12\x94\x01\n\x19GenerateArtifactUploadURL\x12:.continual.rpc.logging.v1.GenerateArtifactUploadURLRequest\x1a;.continual.rpc.logging.v1.GenerateArtifactUploadURLResponse\x12o\n\rLogDataChecks\x12..continual.rpc.logging.v1.LogDataChecksRequest\x1a\x16.google.protobuf.Empty\"\x16\x82\xd3\xe4\x93\x02\x10\"\x0e/v1/dataChecks\x12x\n\x0fLogModelVersion\x12\x30.continual.rpc.logging.v1.LogModelVersionRequest\x1a\x16.google.protobuf.Empty\"\x1b\x82\xd3\xe4\x93\x02\x15\"\x13/v1/logModelVersion\x12\x83\x01\n\x12LogBatchPrediction\x12\x33.continual.rpc.logging.v1.LogBatchPredictionRequest\x1a\x16.google.protobuf.Empty\" \x82\xd3\xe4\x93\x02\x1a\"\x18/v1/batchPredictionStateBEZCgithub.com/continualiq/continual/continual/rpc/logging/v1;loggingv1b\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,google_dot_api_dot_annotations__pb2.DESCRIPTOR,google_dot_protobuf_dot_empty__pb2.DESCRIPTOR,continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2.DESCRIPTOR,])




_LOGHEARTBEATREQUEST = _descriptor.Descriptor(
  name='LogHeartbeatRequest',
  full_name='continual.rpc.logging.v1.LogHeartbeatRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='resource', full_name='continual.rpc.logging.v1.LogHeartbeatRequest.resource', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=212,
  serialized_end=251,
)


_LOGEXPERIMENTSREQUEST = _descriptor.Descriptor(
  name='LogExperimentsRequest',
  full_name='continual.rpc.logging.v1.LogExperimentsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_version', full_name='continual.rpc.logging.v1.LogExperimentsRequest.model_version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='experiments', full_name='continual.rpc.logging.v1.LogExperimentsRequest.experiments', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=253,
  serialized_end=361,
)


_LOGMETRICSREQUEST = _descriptor.Descriptor(
  name='LogMetricsRequest',
  full_name='continual.rpc.logging.v1.LogMetricsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_version', full_name='continual.rpc.logging.v1.LogMetricsRequest.model_version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metrics', full_name='continual.rpc.logging.v1.LogMetricsRequest.metrics', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=363,
  serialized_end=468,
)


_LOGEVALUATIONMODELMETRICSREQUEST = _descriptor.Descriptor(
  name='LogEvaluationModelMetricsRequest',
  full_name='continual.rpc.logging.v1.LogEvaluationModelMetricsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_version', full_name='continual.rpc.logging.v1.LogEvaluationModelMetricsRequest.model_version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='evaluation_model_version_metrics', full_name='continual.rpc.logging.v1.LogEvaluationModelMetricsRequest.evaluation_model_version_metrics', index=1,
      number=23, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=471,
  serialized_end=630,
)


_LOGDATASETSTATSREQUEST = _descriptor.Descriptor(
  name='LogDatasetStatsRequest',
  full_name='continual.rpc.logging.v1.LogDatasetStatsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='resource', full_name='continual.rpc.logging.v1.LogDatasetStatsRequest.resource', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dataset_stats', full_name='continual.rpc.logging.v1.LogDatasetStatsRequest.dataset_stats', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=632,
  serialized_end=740,
)


_LOGPREDICTIONSTATSREQUEST = _descriptor.Descriptor(
  name='LogPredictionStatsRequest',
  full_name='continual.rpc.logging.v1.LogPredictionStatsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='batch_prediction', full_name='continual.rpc.logging.v1.LogPredictionStatsRequest.batch_prediction', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='prediction_stats', full_name='continual.rpc.logging.v1.LogPredictionStatsRequest.prediction_stats', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=742,
  serialized_end=864,
)


_GENERATEARTIFACTUPLOADURLREQUEST = _descriptor.Descriptor(
  name='GenerateArtifactUploadURLRequest',
  full_name='continual.rpc.logging.v1.GenerateArtifactUploadURLRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='resource', full_name='continual.rpc.logging.v1.GenerateArtifactUploadURLRequest.resource', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='path', full_name='continual.rpc.logging.v1.GenerateArtifactUploadURLRequest.path', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='continual.rpc.logging.v1.GenerateArtifactUploadURLRequest.type', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mime_type', full_name='continual.rpc.logging.v1.GenerateArtifactUploadURLRequest.mime_type', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metadata', full_name='continual.rpc.logging.v1.GenerateArtifactUploadURLRequest.metadata', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resumable', full_name='continual.rpc.logging.v1.GenerateArtifactUploadURLRequest.resumable', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=867,
  serialized_end=1003,
)


_GENERATEARTIFACTUPLOADURLRESPONSE = _descriptor.Descriptor(
  name='GenerateArtifactUploadURLResponse',
  full_name='continual.rpc.logging.v1.GenerateArtifactUploadURLResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='url', full_name='continual.rpc.logging.v1.GenerateArtifactUploadURLResponse.url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='artifact', full_name='continual.rpc.logging.v1.GenerateArtifactUploadURLResponse.artifact', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1005,
  serialized_end=1110,
)


_LOGDATACHECKSREQUEST = _descriptor.Descriptor(
  name='LogDataChecksRequest',
  full_name='continual.rpc.logging.v1.LogDataChecksRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='resource', full_name='continual.rpc.logging.v1.LogDataChecksRequest.resource', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data_checks', full_name='continual.rpc.logging.v1.LogDataChecksRequest.data_checks', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1112,
  serialized_end=1214,
)


_LOGMODELVERSIONREQUEST = _descriptor.Descriptor(
  name='LogModelVersionRequest',
  full_name='continual.rpc.logging.v1.LogModelVersionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_version', full_name='continual.rpc.logging.v1.LogModelVersionRequest.model_version', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_paths', full_name='continual.rpc.logging.v1.LogModelVersionRequest.update_paths', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1216,
  serialized_end=1328,
)


_LOGBATCHPREDICTIONREQUEST = _descriptor.Descriptor(
  name='LogBatchPredictionRequest',
  full_name='continual.rpc.logging.v1.LogBatchPredictionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='batch_prediction', full_name='continual.rpc.logging.v1.LogBatchPredictionRequest.batch_prediction', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_paths', full_name='continual.rpc.logging.v1.LogBatchPredictionRequest.update_paths', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1330,
  serialized_end=1451,
)


_GETARTIFACTREQUEST = _descriptor.Descriptor(
  name='GetArtifactRequest',
  full_name='continual.rpc.logging.v1.GetArtifactRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.logging.v1.GetArtifactRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1453,
  serialized_end=1487,
)


_BATCHGETARTIFACTSREQUEST = _descriptor.Descriptor(
  name='BatchGetArtifactsRequest',
  full_name='continual.rpc.logging.v1.BatchGetArtifactsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.logging.v1.BatchGetArtifactsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.logging.v1.BatchGetArtifactsRequest.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1489,
  serialized_end=1546,
)


_BATCHGETARTIFACTSRESPONSE = _descriptor.Descriptor(
  name='BatchGetArtifactsResponse',
  full_name='continual.rpc.logging.v1.BatchGetArtifactsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='artifacts', full_name='continual.rpc.logging.v1.BatchGetArtifactsResponse.artifacts', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1548,
  serialized_end=1633,
)


_DELETEARTIFACTREQUEST = _descriptor.Descriptor(
  name='DeleteArtifactRequest',
  full_name='continual.rpc.logging.v1.DeleteArtifactRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.logging.v1.DeleteArtifactRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1635,
  serialized_end=1672,
)

_LOGEXPERIMENTSREQUEST.fields_by_name['experiments'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._EXPERIMENT
_LOGMETRICSREQUEST.fields_by_name['metrics'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._TRAININGMETRICS
_LOGEVALUATIONMODELMETRICSREQUEST.fields_by_name['evaluation_model_version_metrics'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._EVALUATIONMODELVERSIONMETRICS
_LOGDATASETSTATSREQUEST.fields_by_name['dataset_stats'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATASETSTATS
_LOGPREDICTIONSTATSREQUEST.fields_by_name['prediction_stats'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATASETSTATS
_GENERATEARTIFACTUPLOADURLRESPONSE.fields_by_name['artifact'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ARTIFACT
_LOGDATACHECKSREQUEST.fields_by_name['data_checks'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATACHECKS
_LOGMODELVERSIONREQUEST.fields_by_name['model_version'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._MODELVERSION
_LOGBATCHPREDICTIONREQUEST.fields_by_name['batch_prediction'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._BATCHPREDICTION
_BATCHGETARTIFACTSRESPONSE.fields_by_name['artifacts'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ARTIFACT
DESCRIPTOR.message_types_by_name['LogHeartbeatRequest'] = _LOGHEARTBEATREQUEST
DESCRIPTOR.message_types_by_name['LogExperimentsRequest'] = _LOGEXPERIMENTSREQUEST
DESCRIPTOR.message_types_by_name['LogMetricsRequest'] = _LOGMETRICSREQUEST
DESCRIPTOR.message_types_by_name['LogEvaluationModelMetricsRequest'] = _LOGEVALUATIONMODELMETRICSREQUEST
DESCRIPTOR.message_types_by_name['LogDatasetStatsRequest'] = _LOGDATASETSTATSREQUEST
DESCRIPTOR.message_types_by_name['LogPredictionStatsRequest'] = _LOGPREDICTIONSTATSREQUEST
DESCRIPTOR.message_types_by_name['GenerateArtifactUploadURLRequest'] = _GENERATEARTIFACTUPLOADURLREQUEST
DESCRIPTOR.message_types_by_name['GenerateArtifactUploadURLResponse'] = _GENERATEARTIFACTUPLOADURLRESPONSE
DESCRIPTOR.message_types_by_name['LogDataChecksRequest'] = _LOGDATACHECKSREQUEST
DESCRIPTOR.message_types_by_name['LogModelVersionRequest'] = _LOGMODELVERSIONREQUEST
DESCRIPTOR.message_types_by_name['LogBatchPredictionRequest'] = _LOGBATCHPREDICTIONREQUEST
DESCRIPTOR.message_types_by_name['GetArtifactRequest'] = _GETARTIFACTREQUEST
DESCRIPTOR.message_types_by_name['BatchGetArtifactsRequest'] = _BATCHGETARTIFACTSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetArtifactsResponse'] = _BATCHGETARTIFACTSRESPONSE
DESCRIPTOR.message_types_by_name['DeleteArtifactRequest'] = _DELETEARTIFACTREQUEST
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LogHeartbeatRequest = _reflection.GeneratedProtocolMessageType('LogHeartbeatRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGHEARTBEATREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.LogHeartbeatRequest)
  })
_sym_db.RegisterMessage(LogHeartbeatRequest)

LogExperimentsRequest = _reflection.GeneratedProtocolMessageType('LogExperimentsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGEXPERIMENTSREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.LogExperimentsRequest)
  })
_sym_db.RegisterMessage(LogExperimentsRequest)

LogMetricsRequest = _reflection.GeneratedProtocolMessageType('LogMetricsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGMETRICSREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.LogMetricsRequest)
  })
_sym_db.RegisterMessage(LogMetricsRequest)

LogEvaluationModelMetricsRequest = _reflection.GeneratedProtocolMessageType('LogEvaluationModelMetricsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGEVALUATIONMODELMETRICSREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.LogEvaluationModelMetricsRequest)
  })
_sym_db.RegisterMessage(LogEvaluationModelMetricsRequest)

LogDatasetStatsRequest = _reflection.GeneratedProtocolMessageType('LogDatasetStatsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGDATASETSTATSREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.LogDatasetStatsRequest)
  })
_sym_db.RegisterMessage(LogDatasetStatsRequest)

LogPredictionStatsRequest = _reflection.GeneratedProtocolMessageType('LogPredictionStatsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGPREDICTIONSTATSREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.LogPredictionStatsRequest)
  })
_sym_db.RegisterMessage(LogPredictionStatsRequest)

GenerateArtifactUploadURLRequest = _reflection.GeneratedProtocolMessageType('GenerateArtifactUploadURLRequest', (_message.Message,), {
  'DESCRIPTOR' : _GENERATEARTIFACTUPLOADURLREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.GenerateArtifactUploadURLRequest)
  })
_sym_db.RegisterMessage(GenerateArtifactUploadURLRequest)

GenerateArtifactUploadURLResponse = _reflection.GeneratedProtocolMessageType('GenerateArtifactUploadURLResponse', (_message.Message,), {
  'DESCRIPTOR' : _GENERATEARTIFACTUPLOADURLRESPONSE,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.GenerateArtifactUploadURLResponse)
  })
_sym_db.RegisterMessage(GenerateArtifactUploadURLResponse)

LogDataChecksRequest = _reflection.GeneratedProtocolMessageType('LogDataChecksRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGDATACHECKSREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.LogDataChecksRequest)
  })
_sym_db.RegisterMessage(LogDataChecksRequest)

LogModelVersionRequest = _reflection.GeneratedProtocolMessageType('LogModelVersionRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGMODELVERSIONREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.LogModelVersionRequest)
  })
_sym_db.RegisterMessage(LogModelVersionRequest)

LogBatchPredictionRequest = _reflection.GeneratedProtocolMessageType('LogBatchPredictionRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGBATCHPREDICTIONREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.LogBatchPredictionRequest)
  })
_sym_db.RegisterMessage(LogBatchPredictionRequest)

GetArtifactRequest = _reflection.GeneratedProtocolMessageType('GetArtifactRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETARTIFACTREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.GetArtifactRequest)
  })
_sym_db.RegisterMessage(GetArtifactRequest)

BatchGetArtifactsRequest = _reflection.GeneratedProtocolMessageType('BatchGetArtifactsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETARTIFACTSREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.BatchGetArtifactsRequest)
  })
_sym_db.RegisterMessage(BatchGetArtifactsRequest)

BatchGetArtifactsResponse = _reflection.GeneratedProtocolMessageType('BatchGetArtifactsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETARTIFACTSRESPONSE,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.BatchGetArtifactsResponse)
  })
_sym_db.RegisterMessage(BatchGetArtifactsResponse)

DeleteArtifactRequest = _reflection.GeneratedProtocolMessageType('DeleteArtifactRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEARTIFACTREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.DeleteArtifactRequest)
  })
_sym_db.RegisterMessage(DeleteArtifactRequest)


DESCRIPTOR._options = None

_LOGGINGAPI = _descriptor.ServiceDescriptor(
  name='LoggingAPI',
  full_name='continual.rpc.logging.v1.LoggingAPI',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=1675,
  serialized_end=3246,
  methods=[
  _descriptor.MethodDescriptor(
    name='LogHeartbeat',
    full_name='continual.rpc.logging.v1.LoggingAPI.LogHeartbeat',
    index=0,
    containing_service=None,
    input_type=_LOGHEARTBEATREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\017\"\r/v1/heartbeat',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='LogExperiments',
    full_name='continual.rpc.logging.v1.LoggingAPI.LogExperiments',
    index=1,
    containing_service=None,
    input_type=_LOGEXPERIMENTSREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\021\"\017/v1/experiments',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='LogMetrics',
    full_name='continual.rpc.logging.v1.LoggingAPI.LogMetrics',
    index=2,
    containing_service=None,
    input_type=_LOGMETRICSREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\r\"\013/v1/metrics',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='LogEvaluationModelMetrics',
    full_name='continual.rpc.logging.v1.LoggingAPI.LogEvaluationModelMetrics',
    index=3,
    containing_service=None,
    input_type=_LOGEVALUATIONMODELMETRICSREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\027\"\025/v1/evaluationMetrics',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='LogDatasetStats',
    full_name='continual.rpc.logging.v1.LoggingAPI.LogDatasetStats',
    index=4,
    containing_service=None,
    input_type=_LOGDATASETSTATSREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\022\"\020/v1/datasetStats',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='LogPredictionStats',
    full_name='continual.rpc.logging.v1.LoggingAPI.LogPredictionStats',
    index=5,
    containing_service=None,
    input_type=_LOGPREDICTIONSTATSREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\025\"\023/v1/predictionStats',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetArtifact',
    full_name='continual.rpc.logging.v1.LoggingAPI.GetArtifact',
    index=6,
    containing_service=None,
    input_type=_GETARTIFACTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ARTIFACT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetArtifacts',
    full_name='continual.rpc.logging.v1.LoggingAPI.BatchGetArtifacts',
    index=7,
    containing_service=None,
    input_type=_BATCHGETARTIFACTSREQUEST,
    output_type=_BATCHGETARTIFACTSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteArtifact',
    full_name='continual.rpc.logging.v1.LoggingAPI.DeleteArtifact',
    index=8,
    containing_service=None,
    input_type=_DELETEARTIFACTREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GenerateArtifactUploadURL',
    full_name='continual.rpc.logging.v1.LoggingAPI.GenerateArtifactUploadURL',
    index=9,
    containing_service=None,
    input_type=_GENERATEARTIFACTUPLOADURLREQUEST,
    output_type=_GENERATEARTIFACTUPLOADURLRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='LogDataChecks',
    full_name='continual.rpc.logging.v1.LoggingAPI.LogDataChecks',
    index=10,
    containing_service=None,
    input_type=_LOGDATACHECKSREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\020\"\016/v1/dataChecks',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='LogModelVersion',
    full_name='continual.rpc.logging.v1.LoggingAPI.LogModelVersion',
    index=11,
    containing_service=None,
    input_type=_LOGMODELVERSIONREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\025\"\023/v1/logModelVersion',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='LogBatchPrediction',
    full_name='continual.rpc.logging.v1.LoggingAPI.LogBatchPrediction',
    index=12,
    containing_service=None,
    input_type=_LOGBATCHPREDICTIONREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\032\"\030/v1/batchPredictionState',
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_LOGGINGAPI)

DESCRIPTOR.services_by_name['LoggingAPI'] = _LOGGINGAPI

# @@protoc_insertion_point(module_scope)



import logging
import traceback
import pandas as pd
import json
from datetime import datetime
from dateutil.parser import parse

import uuid
import grpc

from google.protobuf import empty_pb2
from google.protobuf.timestamp_pb2 import Timestamp
from google.protobuf.json_format import MessageToDict

from continual.python.utils import config
from continual.python.sdk.exceptions import InvalidArgumentError
from continual.rpc.dataingest.v1 import dataingest_pb2
from continual.rpc.dataingest.v1 import dataingest_pb2_grpc

from continual.rpc.management.v1 import management_types_pb2
from continual.services.dataingest.gcs_source import FeatureSourceCloudStore
from continual.python.sdk.features.schema_utils import get_type_enum
from continual.python.sdk.features.featurestore import get_featurestore
from continual.python.sdk.query import Query
from continual.python.sdk.datasets import Dataset

ProtoTypeMap = {
    "NUMBER": management_types_pb2.FieldType.NUMBER,
    "TEXT": management_types_pb2.FieldType.TEXT,
    "BOOLEAN": management_types_pb2.FieldType.BOOLEAN,
    "CATEGORICAL": management_types_pb2.FieldType.CATEGORICAL,
    "TIMESTAMP": management_types_pb2.FieldType.TIMESTAMP,
    "DATE": management_types_pb2.FieldType.DATE,
}

ERROR_NO_ID = "Row missing id column."


def collect_fields(qd, coll):
    for f in qd["fields"]:
        coll[f["qualified_name"]] = f["type"]
        if "relation" in f:
            collect_fields(f["relation"], coll)

    if "related" in qd:
        for r in qd["related"]:
            collect_fields(r, coll)


class DataIngestService(dataingest_pb2_grpc.DataIngestAPIServicer):
    def __init__(self):
        self._cfg = config.get_config("dataingest")
        self.jobs = {}

    def cfg(self, context: grpc.ServicerContext = None) -> dict:
        if not context:
            return self._cfg or dict()

        cfg = dict()
        cfg.update(self._cfg or dict())

        context_dict = dict(context.invocation_metadata() or {})
        cfg.update(context_dict or dict())

        return cfg

    def _build_feature_to_proto(self, df, feature_map):
        """
        match DF to schema
        """
        all_records = []
        for key in df.columns:
            if key not in feature_map:
                continue

            series = df[key]
            if len(series.shape) > 1:
                series = df[key].iloc[:, 0]

            if feature_map[key] == "TEXT" or feature_map[key] == "categorical":
                df[key] = series.astype(str, errors="ignore")
            if feature_map[key] == "NUMBER":
                df[key] = pd.to_numeric(series, errors="coerce")
            if feature_map[key] == "BOOLEAN":
                df[key] = series.astype(bool, errors="ignore")
            if feature_map[key] == "TIMESTAMP":
                df[key] = series.dt.to_pydatetime()

        record_dict = df.to_dict(orient="record")

        error_rows = 0
        for r in record_dict:
            try:
                features_record = json.dumps(r, default=str)
                all_records.append(features_record)
            except Exception:
                error_rows = error_rows + 1
                traceback.print_exc()
                continue

        return all_records

    def GetInferredSchema(self, request, context):
        logging.debug("Got GetInferredSchema")

        source_cfg = management_types_pb2.SourceConfig(
            query=request.query, file=request.file
        )

        sample_row_count = request.sample_row_count
        if sample_row_count is None:
            sample_row_count = 100

        try:
            source = None
            if request.table_name is not None and request.table_name != "":
                source = get_featurestore(
                    request.environment.data_store,
                    cfg=self.cfg(context),
                )
                inferred_schema, all_records = source.infer_schema(
                    f"SELECT * FROM {request.table_name}",
                    sample_row_count=sample_row_count,
                    allow_invalid_types=request.allow_invalid_types,
                )

            elif request.query is not None and request.query != "":
                source = get_featurestore(
                    request.environment.data_store,
                    cfg=self.cfg(context),
                )
                inferred_schema, all_records = source.infer_schema(
                    request.query,
                    sample_row_count=sample_row_count,
                    allow_invalid_types=request.allow_invalid_types,
                )
            else:
                uid = str(uuid.uuid1())
                logging.info(
                    "for project "
                    + request.project.name
                    + " attempt to download "
                    + request.file
                )
                source = FeatureSourceCloudStore(uid, None)
                inferred_schema, all_records = source.infer_schema(
                    request.file,
                    sample_row_count=sample_row_count,
                    allow_invalid_types=request.allow_invalid_types,
                )

            features = []

            for sc in inferred_schema:
                fs = management_types_pb2.ColumnConfig(
                    name=sc,
                    type=inferred_schema[sc]["type"],
                    dtype=inferred_schema[sc]["dtype"],
                )
                features.append(fs)

            resp = dataingest_pb2.GetInferredSchemaResponse(
                columns=features,
                record_count=len(all_records),
                values=all_records,
            )

            return resp

        except Exception as e:
            traceback.print_exc()
            context.set_details(str(e))
            context.set_code(grpc.StatusCode.INTERNAL)
            return dataingest_pb2.GetInferredSchemaResponse()

    def GetData(self, request, context):
        logging.debug("Got GetData")
        fs = get_featurestore(
            request.data_store,
            cfg=self.cfg(context),
        )

        feature_map = {}
        index = None
        index_type = None
        for sc in request.columns:
            feature_map[sc.name] = sc.dtype
            if sc.type == management_types_pb2.FieldType.INDEX:
                index = sc
                index_type = sc.type

        if index is None:
            context.set_details("Invalid featureset, no index present")
            context.set_code(grpc.StatusCode.INTERNAL)
            return dataingest_pb2.GetDataResponse()

        token = request.page_token
        if index_type == ProtoTypeMap["NUMBER"]:
            token = float(request.page_token)
        elif index_type == ProtoTypeMap["TIMESTAMP"]:
            token = parse(request.page_token)
        else:
            if token == "":
                token = None

        df = fs.fetch_feature_data(
            name=request.name,
            index=index.name,
            page_size=request.page_size,
            page_token=token,
            table_name=request.table_name,
        )

        next_token = ""
        all_records = []
        if df.shape[0] != 0:
            next_token = str(df[index.name].values[-1])
            all_records = self._build_feature_to_proto(df, feature_map)

        resp = dataingest_pb2.GetDataResponse(
            record_count=len(all_records),
            next_page_token=next_token,
            value=all_records,
        )

        return resp

    def FormQuery(self, request, context):
        fs_map = {}
        for fs in request.feature_sets:
            fs_map[fs.name] = fs

        q = Query.form_query(
            model=request.model,
            all_featuresets=fs_map,
            metadata=request.add_metadata,
        )

        logging.debug(q)
        return q.to_proto()

    def GetServingData(self, request, context):
        logging.debug("Got GetServingData")
        fs = get_featurestore(
            request.data_store,
            cfg=self.cfg(context),
        )
        q_dict = MessageToDict(request.query, preserving_proto_field_name=True)
        overrides = json.loads(request.overrides)

        feature_map = {}
        collect_fields(q_dict, feature_map)

        query = Query(request.query)
        df = fs.fetch_serving_data(query, overrides)
        all_records = self._build_feature_to_proto(df, feature_map)
        resp = dataingest_pb2.GetServingDataResponse(
            value=all_records,
        )
        return resp

    def GetTrainingData(self, request, context):
        logging.debug("Got GetTrainingData")
        fs = get_featurestore(
            request.data_store,
            cfg=self.cfg(context),
        )
        q_dict = MessageToDict(request.query, preserving_proto_field_name=True)

        feature_map = {}
        collect_fields(q_dict, feature_map)

        all_records = []
        query = Query(request.query)
        df_iter = fs.fetch_training_data(query, n=3000)
        try:
            if isinstance(df_iter, pd.DataFrame):
                all_records.extend(self._build_feature_to_proto(df_iter, feature_map))
            else:
                for df in df_iter:
                    all_records.extend(self._build_feature_to_proto(df, feature_map))

                    # TODO: For now we only return first batch.
                    break
        except StopIteration:
            pass

        resp = dataingest_pb2.GetTrainingDataResponse(
            value=all_records,
        )
        return resp

    def ComputeStats(self, request, context):
        logging.debug("ComputeStats: ")
        fstore = get_featurestore(
            request.environment.data_store,
            cfg=self.cfg(context),
        )

        num_rows, computed_at, features = fstore.compute_feature_stats(
            request.name, request.columns, request.table_name
        )
        pb_features = []
        for f in features:
            for key, item in f.items():
                if isinstance(item, datetime):
                    dt = Timestamp()
                    dt.FromDatetime(item)
                    f[key] = dt
            f["type"] = get_type_enum(f.get("type", None))

            pb_features.append(management_types_pb2.ColumnStats(**f))

        timestamp = Timestamp()
        timestamp.FromDatetime(computed_at)
        fss = management_types_pb2.ColumnSetStats(
            name=request.name,
            row_count=num_rows,
            computed_at=timestamp,
            columns=pb_features,
        )
        resp = dataingest_pb2.ComputeStatsResponse(column_set_stats=fss)
        return resp

    def CreateFeatureSetTables(self, request, context):
        logging.debug("CreateFeatureSetTables: ")
        fstore = get_featurestore(
            request.environment.data_store,
            cfg=self.cfg(context),
        )

        fstore.create_feature_table(request.feature_set)
        return empty_pb2.Empty()

    def CreateModelTables(self, request, context):
        logging.debug("CreateModelTables: ")
        fstore = get_featurestore(
            request.environment.data_store,
            cfg=self.cfg(context),
        )

        fs_map = {}
        for fs in request.feature_sets:
            fs_map[fs.name] = fs

        problem_type = Dataset(
            model=request.model,
            featureset_map=fs_map,
            data_store=request.environment.data_store,
        ).inferred_problem_type

        fstore.create_model_tables(
            request.model,
            fs_map,
            problem_type,
            request.prev_index,
            request.prev_time_index,
        )
        return empty_pb2.Empty()

    def CreateFeatureStore(self, request, context):
        logging.debug("CreateFeatureStore: ")
        fstore = get_featurestore(
            request.environment.data_store, cfg=self.cfg(context), init=False
        )
        splits = request.environment.name.split("/")
        if len(splits) != 2:
            raise InvalidArgumentError(
                "Project name [{}] incorrect".format(request.environment.name)
            )

        fstore.create_project_schema()
        return empty_pb2.Empty()

    def CleanupFeatureTables(self, request, context):
        logging.debug("CleanupFeatureTablesRequest: ")
        fstore = get_featurestore(
            request.environment.data_store,
            cfg=self.cfg(context),
        )

        if "featureSet" in request.name:
            fstore.drop_featureset_view(request.name)
        elif "models" in request.name:
            fstore.drop_model_views(request.name)

        return empty_pb2.Empty()

    def DeleteFeatureStore(self, request, context):
        logging.debug("DeleteFeatureStore: ")
        fstore = get_featurestore(
            request.environment.data_store,
            cfg=self.cfg(context),
        )
        splits = request.environment.name.split("/")
        if len(splits) != 2:
            raise InvalidArgumentError(
                "Project name [{}] incorrect".format(request.environment.name)
            )

        fstore.delete_project_schema(fstore.db_schema)
        return empty_pb2.Empty()

    def BrowseDataWarehouse(self, request, context):
        logging.debug("BrowseDataWarehouse: ")

        fstore = get_featurestore(request.data_store, cfg=self.cfg(context), init=False)
        resp_data = []
        type_map = None
        if request.object_type == dataingest_pb2.BrowseDataWarehouseRequest.DATABASE:
            resp_data = fstore.get_database_names()
        elif request.object_type == dataingest_pb2.BrowseDataWarehouseRequest.SCHEMA:
            resp_data = fstore.get_schema_names(request.database)
        elif request.object_type == dataingest_pb2.BrowseDataWarehouseRequest.TABLE:
            resp_data = fstore.get_table_names(request.database, request.schema)
        elif request.object_type == dataingest_pb2.BrowseDataWarehouseRequest.COLUMN:
            resp_data, type_map = fstore.get_column_names(
                request.database, request.schema, request.table
            )

        return dataingest_pb2.BrowseDataWarehouseResponse(
            names=resp_data, types=type_map
        )

    def Seed(self, request, context):
        logging.debug("Seed: ")
        fstore = get_featurestore(request.environment.data_store, cfg=self.cfg(context))

        table_name = fstore.load_data(
            request.file_path, request.table_name, replace=True
        )
        return dataingest_pb2.SeedResponse(
            schema_name=fstore.db_schema, table_name=table_name
        )



import sqlalchemy as db
import logging
import urllib
import traceback
from datetime import datetime
import pandas as pd
from sqlalchemy.exc import IntegrityError, ProgrammingError

from .prediction_store import PredictionStore, chunk_it
from .featurestore import FeatureStore, Typemap
from continual.python.sdk.exceptions import InvalidArgumentError
from continual.rpc.management.v1 import management_types_pb2
from .query_templates import stats_template_mssql


def create_connection(data_store):
    """
    Create mssql connection given data_store.
    """
    connection_string = "Driver={driver};Server=tcp:{host},{port};Database={database};Uid={user};Pwd={password};Encrypt=yes;TrustServerCertificate=no;Connection Timeout=30;".format(
        driver="{ODBC Driver 17 for SQL Server}",
        host=data_store.host,
        port=data_store.port,
        database=data_store.database,
        user=data_store.username,
        password=data_store.password,
    )

    params = urllib.parse.quote_plus(connection_string)
    engine = db.create_engine("mssql+pyodbc:///?odbc_connect=%s" % params)
    connection = engine.connect()
    return engine, connection


class MSSqlFeatureStore(FeatureStore, PredictionStore):

    """
    MSSqlConnector implements both source of features and featurestore functions.

    Currently sources are not pooled. In a larger tenant with frequent and multiple fetches,
    we would need to consider connection pooling.
    """

    def __init__(self, data_store, source, cfg):
        self.data_store = data_store
        self.source = source
        self.corpus_id = None

        try:
            self.engine, self.connection = create_connection(data_store)
        except Exception:
            logging.debug(traceback.format_exc())
            raise InvalidArgumentError(
                "Connection to mssql failed, please check connection settings.",
                details=None,
            )
        super(MSSqlFeatureStore, self).__init__()

    def cleanup(self):
        tab_query = "select table_name, table_type from information_schema.tables where table_schema='{}'".format(
            self.db_schema
        )
        results = self.connection.execute(tab_query)
        views = []
        tables = []
        for row in results:
            if row[1] == "VIEW":
                views.append(row[0])
            else:
                tables.append(row[0])

        if len(tables) == 0:
            return

        # TODO: fix when/if we support mssqlserver
        logging.error("cleanup not supported for mlsqlserver")
        # with self.connection.begin():
        #     for table_name in tables:
        #         drop_table_query = "DROP TABLE {}.{}".format(project, table_name)
        #         self.connection.execute(drop_table_query)

        #     for table_name in views:
        #         drop_table_query = "DROP VIEW {}.{}".format(project, table_name)
        #         self.connection.execute(drop_table_query)

        #     drop_schema_query = "DROP SCHEMA {}".format(project)
        #     self.connection.execute(drop_schema_query)
        return

    def create_feature_table(self, feature_set):
        """
         Creates featuretable.
         This does not check for the existance of the featuretable and should
           be done prior to calling this.

        Args: Feature set for which the tables are to be created.≈
        Returns None
        """
        for f in feature_set.schema.metadata:
            if (
                f.type == management_types_pb2.FieldType.NUMBER_LIST
                or f.type == management_types_pb2.FieldType.TEXT_LIST
            ):
                # Skip array types for mssql for now.
                return

        for f in feature_set.schema.features:
            if (
                f.type == management_types_pb2.FieldType.NUMBER_LIST
                or f.type == management_types_pb2.FieldType.TEXT_LIST
            ):
                # Skip array types for mssql for now.
                return

        for f in feature_set.schema.entities:
            if (
                f.type == management_types_pb2.FieldType.NUMBER_LIST
                or f.type == management_types_pb2.FieldType.TEXT_LIST
            ):
                # Skip array types for mssql for now.
                return

        for f in feature_set.schema.models:
            if (
                f.type == management_types_pb2.FieldType.NUMBER_LIST
                or f.type == management_types_pb2.FieldType.TEXT_LIST
            ):
                # Skip array types for mssql for now.
                return

        return super().create_feature_table(feature_set)

    def materialize_data(self, query):
        logging.info("Going to execute query " + query)
        results = self.connection.execute(query).fetchone()
        logging.info(results[0])
        return results[0]

    def fetch_data(self, page_start, page_size=100000, schema=None):
        # Not implemented
        return None

    def close(self):
        self.connection.close()

    def create_project_schema(self):
        results = self.connection.execute(
            "select count(*) from sys.schemas where name='{}'".format(self.db_schema)
        ).fetchone()
        if results[0] > 0:
            return
        self.connection.execute("CREATE SCHEMA {}".format(self.db_schema))

    def create_featureset_view(self, feature_set):
        project, table_name = self.get_table_name(feature_set)
        table_name = "featureset_{}".format(table_name)
        self.drop_table(table_name, "")

        query = feature_set.schema.source.query.strip().rstrip(";")
        self.connection.execute(
            "CREATE VIEW {}.{} AS ({})".format(project, table_name, query)
        )
        return

    def drop_table(self, table_name, view_name):
        archive_table_name = table_name + "_" + datetime.now().strftime("%Y%m%d%H%m%S")
        trans = self.connection.begin()
        try:
            self.connection.execute(
                """IF EXISTS (SELECT * from INFORMATION_SCHEMA.TABLES WHERE 
                                    table_schema='{}' and table_name='{}') 
                        BEGIN
                            EXEC sp_rename '{}', '{}'
                        END
                """.format(
                    self.db_schema,
                    table_name,
                    self.db_schema + "." + table_name,
                    archive_table_name,
                )
            )

            trans.commit()
        except:
            print(traceback.format_exc())
            logging.debug(traceback.format_exc())
            trans.rollback()
            raise

    def compute_feature_stats(self, featureset):
        proj, table_name = self.get_table_name(featureset)

        num_rows = 0
        features = []
        res = self.connection.execute(
            "SELECT count(*) as num_rows FROM {}.{}".format(proj, table_name)
        )
        row = next(res)
        num_rows = row[0]

        field_array = []  # TODO
        for f in featureset.schema.metadata:
            field_array.append(f)

        for f in featureset.schema.features:
            field_array.append(f)

        for f in featureset.schema.entities:
            field_array.append(f)

        for f in featureset.schema.models:
            field_array.append(f)

        params = {
            "SchemaName": featureset.name,
            "AllFeatures": field_array,
            "FullName": proj + "." + table_name,
            "Entity": featureset.schema.name,
        }

        stats_query = stats_template_mssql.render(params=params)
        cur = self.connection.execute(stats_query)
        done = False
        computed_at = None
        while not done:
            try:
                row = next(cur)
                d = dict(zip(row.keys(), row))
                if "computed_at" in d:
                    computed_at = d["computed_at"]
                    del d["computed_at"]

                features.append(d)
            except StopIteration:
                done = True

        return num_rows, computed_at, features

    def get_type(self) -> str:
        return "sqlserver"

    def is_hosted(self):
        return False

    def create_prediction_model_table(
        self, feature_set, model_id, prev_index, prev_time_index
    ):
        """
         Creates featuretable.
         This does not check for the existance of the featuretable and should
           be done prior to calling this.

        Args: Feature set for which the tables are to be created.≈
        Returns None
        """
        project, table_name = self.get_table_name(feature_set)
        meta = db.MetaData(schema=project)

        index = None
        time_index = None
        for f in feature_set.schema.metadata:
            if f.index:
                index = f
            elif f.time_index:
                time_index = f

        _, model_pred_name = self.get_model_prediction_table_name(feature_set)

        columns = []
        columns.append(db.Column("batch_prediction_job", db.Text))
        columns.append(db.Column("model_version", db.Text))
        columns.append(db.Column("prediction_time", db.TIMESTAMP(timezone=True)))
        columns.append(db.Column(index.name, Typemap[index.type]))
        if time_index is not None:
            columns.append(db.Column(time_index.name, Typemap[time_index.type]))
        columns.append(db.Column("label", db.Text))
        columns.append(db.Column("category", db.Text))
        columns.append(db.Column("score", db.Float))
        columns.append(db.Column("details", db.Text))

        db.Table(model_pred_name, meta, *columns)

        # Following retry logic is added in case a parallel operation
        # creates a schema that might not have existed at the start of the session.
        count = 0
        done = False
        while not done:
            try:
                meta.create_all(self.engine)
                done = True
            except (IntegrityError, ProgrammingError):
                logging.debug(traceback.format_exc())
                logging.debug("Error while creating feature table " + str(count))
                count = count + 1
                if count == 2:
                    raise

    def get_stats_query_template(self):
        return stats_template_mssql

    def get_database_names(self):
        raise NotImplementedError()

    def get_schema_names(self, database):
        raise NotImplementedError()

    def get_table_names(self, database, schema):
        raise NotImplementedError()

    def get_column_names(self, database, schema, table):
        raise NotImplementedError()

    def write_prediction(
        self,
        df: pd.DataFrame,
        id_column_name: str,
        ts_column_name: str,
        model_query: any,
        output_name: str,
        job_name: str,
        job_start_time: datetime,
        model,
        model_version,
        time_index_dtype,
    ):
        logging.debug("in write_prediction")
        if self.data_store is None:
            logging.debug("skipping write_prediction")
            return

        project_id, pred_table = self.get_model_prediction_table_name(model)
        df = self._process_dataframe(
            df, id_column_name, ts_column_name, job_name, model_version
        )

        ch = chunk_it(0, df.shape[0])
        while ch is not None:
            df_slice = df[ch[0] : ch[1]]
            try:
                df_slice.to_sql(
                    pred_table,
                    self.connection,
                    schema=project_id,
                    if_exists="append",
                    index=False,
                )
            except:
                logging.info(traceback.format_exc()[:3000])
                raise

            ch = chunk_it(ch[1], df.shape[0])

        return 0



import sqlalchemy as db
from sqlalchemy.sql import select
from sqlalchemy.exc import IntegrityError, ProgrammingError
from sqlalchemy import exc

import logging
import pandas as pd
import uuid
import re
from jinja2 import Template
from functools import wraps
import traceback
from json import loads, dumps
from datetime import datetime
import shortuuid

from snowflake.sqlalchemy import VARIANT
from sqlalchemy.sql.schema import MetaData
from snowflake.connector.constants import FIELD_ID_TO_NAME
from snowflake.connector import connect as snow_con
import snowflake.connector.pandas_tools as ptools
from snowflake.connector.errors import ProgrammingError as SFProgrammingError
from continual.python.utils.utils import is_select_query, create_limit_query

from continual.rpc.management.v1 import management_types_pb2
from .featurestore import FeatureStore, Typemap
from .prediction_store import PredictionStore
from continual.python.sdk.exceptions import InvalidArgumentError
from .query_templates import stats_template_snowflake
from continual.services.dataingest.gcs_source import FeatureSourceCloudStore


TABLE_NOT_FOUND_ERROR_MSG = "tables(s) not found. Check if there are double quotes around table or schema names. Snowflake considers quotes names as case sensitve"
insert_template = Template(
    """INSERT INTO  {{param.db_schema}}.{{param.table_name}} ({% for r in param.insert_list %}{%if loop.index0 != 0%},{% endif %}{{r}}{% endfor %})
        SELECT {% for r in param.columns %}{%if loop.index0 != 0%},{% endif %}{{r}}{% endfor %} 
        FROM {{param.db_schema}}.{{param.temp_table}}"""
)

COLUMN_STATS_BATCH = 250


def column_name_mapper(col_name):
    new_name = re.sub("[^a-zA-Z0-9_]", "_", col_name)
    return new_name.lower()


def create_incremental_query(query, columns, page_data):

    """
    A user given query is wrapped to perform incremental data extraction.

    Returns: sqlalchemy query statement object prepared for incremental data pull.
    """
    if columns is None or len(columns) == 0:
        return db.text(query)

    if page_data is None or len(page_data.keys()) == 0:
        return db.text(query)

    all_columns = []
    for col in columns:
        if col in page_data and page_data[col] is not None:
            all_columns.append(db.column(col))

    # Check if are able to do incremental ingest. If not do full ingest
    if len(all_columns) == 0:
        return db.text(query)

    # Prepare user given text query for use as a subquery.
    q2 = db.text(query).columns(*all_columns).alias("itr")

    # Wrap the user given query with column starts.
    # So a query q will be wraped as
    # SELECT * from (<q>) as itr where itr.<col1> > <cdata1> AND....
    q_full = select([db.text("*")])
    for col in columns:

        # Ignore columns for which we do not have start data.
        if col not in page_data or page_data[col] is None:
            continue
        cdata = page_data[col]
        q_full = q_full.where(q2.c[col] > cdata)

    return q_full


def handle_snowflake_exception(func):
    """Decorator that parses Snowflake exceptions raised through SQLAlchemy.

    Args:
        func: Function to wrap.
    Returns:
        A function that will raise an informative exception.
    """

    @wraps(func)
    def wrapped_func(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except SFProgrammingError as e:
            if e.msg.startswith("invalid identifier"):
                raise InvalidArgumentError(
                    message=f"{e.msg}, this can happen if the underlying table in snowflake has case sensitive column names",
                    details=None,
                )
        except exc.SQLAlchemyError as e:
            if isinstance(e, exc.ProgrammingError):
                msg = str(e.orig) or ""
                msg = msg.split("\n")[-1]
                if not msg:
                    msg = str(e.orig)
                if msg.startswith("invalid identifier"):
                    msg = f"{msg}, this can happen if the underlying table in snowflake has case sensitive column names"
                elif msg.startswith(
                    "Object does not exist, or operation cannot be performed"
                ):
                    msg = f"{msg}, Please check if you have sufficient permissions"
                raise InvalidArgumentError(message=msg, details=None)
            elif isinstance(e, exc.InvalidRequestError):
                msg = str(e)
                if "requested table(s) not available in Engine" in msg:
                    msg = TABLE_NOT_FOUND_ERROR_MSG
                raise InvalidArgumentError(message=msg, details=None)
            raise e
        except:
            raise

    return wrapped_func


def create_connection(data_store):
    """
    Create snowflake connection given data_store.
    """
    account = data_store.snowflake.host
    if account.endswith(".snowflakecomputing.com"):
        account = account[: len(account) - len(".snowflakecomputing.com")]

    engine_url = db.engine.URL.create(
        "snowflake",
        username=data_store.snowflake.username,
        password=data_store.snowflake.password,
        host=account,
        database=data_store.snowflake.database,
        query=dict(
            warehouse=data_store.snowflake.warehouse,
            role=data_store.snowflake.role,
            application="Continual",
        ),
    )

    sf_eng = db.create_engine(engine_url)

    connection = sf_eng.connect()
    # connection.execute("alter session set quoted_identifiers_ignore_case = true")
    return sf_eng, connection


class SnowflakeFeatureStore(FeatureStore, PredictionStore):

    """
    SnowflakeFeatureStore implements both source of features and featurestore functions.

    Currently sources are not pooled. In a larger tenant with frequent and multiple fetches,
    we would need to consider connection pooling.
    """

    DTYPE_TO_LOGICAL = {
        "FIXED": "NUMBER",
        "REAL": "NUMBER",
        "TEXT": "CATEGORICAL",
        "DATE": "TIMESTAMP",
        "TIMESTAMP": "TIMESTAMP",
        "VARIANT": "TEXT",
        "TIMESTAMP_LTZ": "TIMESTAMP",
        "TIMESTAMP_TZ": "TIMESTAMP",
        "TIMESTAMP_NTZ": "TIMESTAMP",
        "OBJECT": "TEXT",
        "ARRAY": "STRING_ARRAY",
        "BINARY": "TEXT",
        "TIME": "TIMESTAMP",
        "BOOLEAN": "BOOLEAN",
    }

    def __init__(self, data_store, source, cfg):
        self.data_store = data_store
        self.source = source
        self.corpus_id = None

        try:
            self.engine, self.connection = create_connection(data_store)
        except Exception:
            raise InvalidArgumentError(
                "Connection to Snowflake failed, please check connection settings.",
                details=None,
            )
        super(SnowflakeFeatureStore, self).__init__()

    def cleanup(self):
        print("Dropping ", self.db_schema)
        self.connection.execute("DROP SCHEMA IF EXISTS " + self.db_schema + " CASCADE")
        return

    @handle_snowflake_exception
    def create_project_schema(self):
        return super().create_project_schema()

    @handle_snowflake_exception
    def create_feature_table(self, feature_set):
        """
         Creates featuretable.
         This does not check for the existance of the featuretable and should
           be done prior to calling this.

        Args: Feature set for which the tables are to be created.≈
        Returns None
        """
        return super().create_feature_table(feature_set)

    @handle_snowflake_exception
    def materialize_data(self, query):
        logging.info("Going to execute query " + query)
        results = self.connection.execute(query).fetchone()
        logging.info(results[0])
        return results[0]

    @handle_snowflake_exception
    def fetch_data(self, page_start, page_size=100000, schema=None):
        # TODO: fetch_data would be enhanced for batch gets with additional parameters.
        query = self.source.query.rstrip().rstrip(";")

        if (
            self.source.incremental_column is not None
            and self.source.incremental_column != ""
            and page_start is not None
        ):
            columns = [self.source.incremental_column]
            query = create_incremental_query(query, columns, page_start)

        logging.info(query)

        class DFIterator:
            def __init__(self, connection, query):
                self._connection = connection
                self._query = query
                self._current_index = 0
                self.chunker = None

            def __iter__(self):
                self.chunker = None
                return self

            def __next__(self):
                if self.chunker is not None:
                    return next(self.chunker)

                self.chunker = pd.read_sql(
                    self._query,
                    self._connection,
                    chunksize=page_size,
                )

                self._current_index = self._current_index + 1
                return next(self.chunker)

        return DFIterator(self.connection, query)

    @handle_snowflake_exception
    def infer_schema(
        self,
        query,
        excludes=None,
        sample_row_count=100,
        allow_invalid_types=False,
        timeout_ms=0,
    ):
        inferred_schema = {}
        if is_select_query(query):
            query = create_limit_query(query, sample_row_count)
        logging.info("about to execute query: {}".format(query))
        res = self.connection.execute(query)
        for j in res.cursor.description:
            col = j[0].lower()
            if excludes is not None and col in excludes:
                continue

            dtype = FIELD_ID_TO_NAME[j[1]]
            if allow_invalid_types and dtype not in self.DTYPE_TO_LOGICAL:
                inferred_schema[col] = {"type": "INVALID", "dtype": dtype}
            elif dtype in self.DTYPE_TO_LOGICAL:
                inferred_schema[col] = {
                    "type": self.DTYPE_TO_LOGICAL[dtype],
                    "dtype": dtype,
                }

        all_rows = []
        for row in res:
            drows = dict(row)

            # Filter our columns in the exclude list.
            if excludes is not None:
                for e in excludes:
                    if e in drows:
                        del drows[e]

            all_rows.append(dumps(drows, default=str))
            if len(all_rows) >= sample_row_count:
                break

        return inferred_schema, all_rows

    def set_db_schema(self, db_schema):
        db_schema_name = db_schema
        # if schema is the project name
        splits = db_schema.split("/")
        if len(splits) == 2:
            db_schema_name = splits[1]

        self.db_schema = db_schema_name
        # self.connection.execute("USE schema {}".format(db_schema))

    @handle_snowflake_exception
    def close(self):
        self.connection.close()

    @handle_snowflake_exception
    def drop_table(self, table_name, view_name):
        return super().drop_table(table_name, view_name)

    @handle_snowflake_exception
    def fetch_featureset_data(
        self,
        feature_set,
        index,
        page_size,
        page_token,
    ):
        return super().fetch_featureset_data(feature_set, index, page_size, page_token)

    @handle_snowflake_exception
    def fetch_training_data(
        self,
        q,
        all_instances=False,
        n=None,
        local=False,
    ):
        return super().fetch_training_data(
            q,
            all_instances,
            n,
            local=local,
        )

    @handle_snowflake_exception
    def fetch_serving_data(self, q, overrides, data_store=None):
        return super().fetch_serving_data(q, overrides=overrides, data_store=data_store)

    def get_type(self) -> str:
        return "snowflake"

    def is_hosted(self):
        return False

    def get_stats_query_template(self):
        return stats_template_snowflake

    @handle_snowflake_exception
    def _create_prediction_model_table(
        self, model, model_pred_name, meta: MetaData, problem_type
    ):
        """
         Creates featuretable.
         This does not check for the existance of the featuretable and should
           be done prior to calling this.

        Args: Feature set for which the tables are to be created.≈
        Returns None
        """
        logging.info("snowflake: _create_prediction_model_table")
        index = None
        time_index = None
        target_column = model.schema.target
        target_column_dtype = db.TEXT
        for f in model.schema.columns:
            if (
                f.type == management_types_pb2.FieldType.INDEX
                or f.name == model.schema.index
            ):
                index = f
            elif (
                f.type == management_types_pb2.FieldType.TIME_INDEX
                or f.name == model.schema.time_index
            ):
                time_index = f
            elif f.name == target_column:
                target_column_dtype = Typemap[f.dtype]

        columns = []
        columns.append(db.Column("batch_prediction", db.Text))
        columns.append(db.Column("model_version", db.Text))
        columns.append(db.Column("prediction_time", db.TIMESTAMP(timezone=True)))
        columns.append(db.Column(f"features", VARIANT))

        if index is not None:
            columns.append(db.Column(index.name, Typemap[index.dtype]))

        if time_index is not None:
            columns.append(db.Column(time_index.name, Typemap[time_index.dtype]))
        columns.append(db.Column(f"{target_column}_prediction", target_column_dtype))

        if problem_type == "multiclass_classification":
            columns.append(db.Column(f"{target_column}_prediction_details", VARIANT))
            columns.append(db.Column(f"{target_column}_prediction_score", db.Float))
        if problem_type == "binary_classification":
            columns.append(
                db.Column(f"{target_column}_true_prediction_score", db.Float)
            )

        db.Table(model_pred_name, meta, *columns)

        # Following retry logic is added in case a parallel operation
        # creates a schema that might not have existed at the start of the session.
        count = 0
        done = False
        while not done:
            try:
                meta.create_all(self.engine)
                done = True
            except (IntegrityError, ProgrammingError):
                logging.debug(traceback.format_exc())
                logging.debug("Error while creating feature table " + str(count))
                count = count + 1
                if count == 2:
                    raise

    @handle_snowflake_exception
    def get_database_names(self):
        databases = []
        result_proxy = self.connection.execute("SHOW DATABASES")
        for rowproxy in result_proxy:
            d = {}
            for col, val in rowproxy.items():
                d[col] = val
            databases.append(d["name"])

        return databases

    @handle_snowflake_exception
    def get_schema_names(self, database):
        schemas = []
        result_proxy = self.connection.execute(
            "SHOW SCHEMAS IN DATABASE {}".format(database)
        )
        for rowproxy in result_proxy:
            d = {}
            for col, val in rowproxy.items():
                d[col] = val
            if d["name"] == "INFORMATION_SCHEMA":
                continue
            schemas.append(d["name"])

        return schemas

    @handle_snowflake_exception
    def get_table_names(self, database, schema, views=True):
        tables = []
        result_proxy = self.connection.execute(
            "SHOW TABLES IN SCHEMA {}.{}".format(database, schema)
        )
        for rowproxy in result_proxy:
            d = {}
            for col, val in rowproxy.items():
                d[col] = val
            tables.append(d["name"])

        if views:
            result_proxy = self.connection.execute(
                "SHOW VIEWS IN SCHEMA {}.{}".format(database, schema)
            )
            for rowproxy in result_proxy:
                d = {}
                for col, val in rowproxy.items():
                    d[col] = val
                tables.append(d["name"])
        return tables

    @handle_snowflake_exception
    def get_column_names(self, database, schema, table):
        columns = []
        type_map = {}
        result_proxy = self.connection.execute(
            "SHOW COLUMNS IN TABLE {}.{}.{}".format(database, schema, table)
        )
        for rowproxy in result_proxy:
            d = {}
            for col, val in rowproxy.items():
                d[col] = val
            columns.append(d["column_name"])
            if "data_type" in d:
                type_dict = loads(d["data_type"])
                type_map[d["column_name"]] = type_dict["type"]

        return columns, type_map

    @handle_snowflake_exception
    def compute_feature_stats(self, name, columns, table_name=None):
        if len(columns) <= COLUMN_STATS_BATCH:
            return super().compute_feature_stats(name, columns, table_name)

        num_rows = None
        computed_at = 0
        all_features = []
        i = 0
        j = COLUMN_STATS_BATCH
        while i < len(columns):
            logging.info("I {} J {}".format(i, j - 1))
            if num_rows is None:
                num_rows, computed_at, features = super().compute_feature_stats(
                    name, columns[i : j - 1], table_name
                )
            else:
                _, _, features = super().compute_feature_stats(
                    name, columns[i : j - 1], table_name
                )
            all_features.extend(features)
            i = j
            if j + COLUMN_STATS_BATCH >= len(columns):
                j = j + (len(columns) - j)
            else:
                j = j + COLUMN_STATS_BATCH
            if i == j:
                break

        logging.info("All-Features{} I {} J {}".format(len(all_features), i, j - 1))
        return num_rows, computed_at, all_features

    @handle_snowflake_exception
    def load_data(self, filename, table_name, replace=False):
        source = FeatureSourceCloudStore(str(uuid.uuid4()), filename)
        df = source.fetch_all()

        df = df.rename(column_name_mapper, axis=1)
        table_name = table_name.lower()
        df.head(0).to_sql(
            name=table_name,
            con=self.connection,
            schema=self.db_schema,
            if_exists="replace",
            index=False,
        )

        account = self.data_store.snowflake.host
        if account.endswith(".snowflakecomputing.com"):
            account = account[: len(account) - len(".snowflakecomputing.com")]

        with snow_con(
            account=account,
            user=self.data_store.snowflake.username,
            password=self.data_store.snowflake.password,
            database=self.data_store.snowflake.database,
            warehouse=self.data_store.snowflake.warehouse,
            role=self.data_store.snowflake.role,
            application="Continual",
        ) as con:
            con.cursor().execute(f"Use schema {self.db_schema}")
            ptools.write_pandas(
                con,
                df,
                database=self.data_store.snowflake.database,
                schema=self.db_schema,
                table_name=table_name,
                quote_identifiers=False,
            )
        return table_name

    def write_prediction(
        self,
        df: pd.DataFrame,
        id_column_name: str,
        ts_column_name: str,
        model_query: any,
        output_name: str,
        job_name: str,
        job_start_time: datetime,
        model,
        model_version,
        time_index_dtype,
    ):
        logging.debug("in write_prediction")
        if self.data_store is None:
            logging.debug("skipping write_prediction")
            return

        project_id, pred_table = self.get_model_prediction_table_name(model)
        df = self._process_dataframe(
            df, id_column_name, ts_column_name, job_name, model_version
        )
        df["prediction_time"] = job_start_time.strftime("%Y-%m-%d %H:%M:%S")

        temp_table_name = pred_table + "_" + shortuuid.uuid().lower()
        ## Insert into temp table.
        with snow_con(
            user=self.data_store.snowflake.username,
            password=self.data_store.snowflake.password,
            account=self.data_store.snowflake.host,
            database=self.data_store.snowflake.database,
            warehouse=self.data_store.snowflake.warehouse,
            role=self.data_store.snowflake.role,
            application="Continual",
        ) as con:
            con.cursor().execute(f"Use schema {self.db_schema}")
            con.cursor().execute(
                f"CREATE TEMP TABLE {self.db_schema}.{temp_table_name} LIKE {self.db_schema}.{pred_table}"
            )

            ptools.write_pandas(
                con,
                df,
                database=self.data_store.snowflake.database,
                schema=self.db_schema,
                table_name=temp_table_name,
                quote_identifiers=False,
            )

            all_columns = []
            insert_list = []

            for column in df.columns:
                insert_list.append(column)
                if column == "features":
                    all_columns.append("PARSE_JSON(features) as features")
                elif column.endswith("_details"):
                    all_columns.append(f"PARSE_JSON({column}) as {column}")
                else:
                    all_columns.append(column)

            logging.info("all columns: {}".format(all_columns))
            param = {
                "db_schema": self.db_schema,
                "temp_table": temp_table_name,
                "table_name": pred_table,
                "columns": all_columns,
                "insert_list": insert_list,
            }
            query = insert_template.render(param=param)
            logging.info("query: {}".format(query))
            con.cursor().execute(query)

            # get prediction count
            count_query = f"SELECT COUNT(1) from {self.db_schema}.{pred_table} WHERE batch_prediction = %s"
            count_cur = con.cursor()
            count_cur.execute(count_query, (job_name))
            res = count_cur.fetchone()
            return res[0]



import mlflow.pyfunc
import cloudpickle
from pandas import DataFrame
import os
import numpy as np


class ContinualSavedModel:
    def __init__(self, model, prediction_model):
        self.model = model
        self.prediction_model = prediction_model

    def predict_proba(self, df):
        return self.model.predict_proba(
            df, model=self.prediction_model, as_pandas=False, as_multiclass=False
        )

    def predict(self, df):
        return self.model.predict(df, model=self.prediction_model)


class ContinualSavedModelWrapper(mlflow.pyfunc.PythonModel):
    """
    Class to wrap continual models
    """

    def load_context(self, context):
        """This method is called when loading an MLflow model with pyfunc.load_model(), as soon as the Python Model is constructed.
        Args:
            context: MLflow context where the model artifact is stored.
        """
        file_to_read = open(context.artifacts["continual_model"], "rb")
        os.chdir(context.artifacts["model_files"])
        loaded_object = cloudpickle.load(file_to_read)
        file_to_read.close()
        self.model = loaded_object

    def predict(self, context, model_input):
        """This is an abstract function. We customized it into a method to fetch the FastText model.
        Args:
            context ([type]): MLflow context where the model artifact is stored.
            model_input ([type]): the input data to fit into the model.
        Returns:
            [type]: the loaded model artifact.
        """
        print("context: {}".format(context))
        results_predict = self.model.predict(model_input)
        results_proba = self.model.predict_proba(model_input)

        result_df = DataFrame(
            {
                "predict": results_predict,
                "predict_proba": None,
            }
        )
        result_df.reset_index(inplace=True)
        if isinstance(results_proba, np.ndarray):
            result_df["predict_proba"] = results_proba.tolist()
        else:
            result_df["predict_proba"] = results_proba

        return result_df



from __future__ import annotations
from typing import Optional


class Manager:
    """Base class for resource managers"""

    name_pattern: str
    parent: str
    client: client.Client

    def __init__(self, client: client.Client, parent: str = "") -> None:
        self.client = client
        self.parent = parent

    def name(self, id: str, parent: Optional[str] = None) -> str:
        """Generates a resource name from parent and id.

        Adds wildcard for any missing parent elements.

        Arguments:
            id: Name or id of resource.
            parent: (Optional) Override parent name.
        """
        if "/" in id:
            # Don't allow names to override manager parent config since this is confusing
            # and is typically a bug in the user code.
            if parent is not None and parent != "" and not id.startswith(parent):
                raise ValueError(f"Resource {id} not a child of {parent}.")
            return id
        pattern_parts = self.name_pattern.split("/")
        parent = parent or self.parent or ""
        parent_parts = parent.split("/")
        out = []
        for i in range(len(pattern_parts)):
            if (i + 1) % 2 == 0:
                if i < len(parent_parts):
                    out.append(parent_parts[i])
                else:
                    out.append("-")
            else:
                out.append(pattern_parts[i])
        out[-1] = id
        return "/".join(out)



from __future__ import annotations
import os
import yaml

from typing import List, Union, Optional, Iterator
from continual.python.sdk.exceptions import FailedError
from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.waiter import wait
from continual.python.sdk.iterators import Pager
from continual.python.sdk.events import EventManager


class ChangeManager(Manager):
    """Manages change workflows"""

    name_pattern: str = "projects/{project}/changes/{change}"

    def push(
        self,
        paths: Union[List[str], str],
        plan_only=False,
        purge=False,
        train_all=False,
        train_none=False,
        profile_none=False,
        update_all=False,
        message=None,
        trigger="SDK",
        yaml_text=None,
        resource="",
    ) -> Change:
        """Change a directory of schemas to Continual.

        Change recursively walks a directory and uploads all yaml
        schema files found.

        Args:
            paths: Files or directories containing schemas.

        Returns:
            A list of feature sets.
        """

        all_files = []
        if isinstance(paths, str):
            paths = [paths]

        for path in paths:
            if os.path.isdir(path):
                for root, dirs, rfiles in os.walk(path, topdown=False):
                    for name in rfiles:
                        if name.endswith(".yaml") or name.endswith(".yml"):
                            all_files.append(os.path.join(root, name))
            elif (
                os.path.isfile(path) and path.endswith(".yaml") or path.endswith(".yml")
            ):
                all_files.append(path)
            elif not resource:
                raise ValueError("path must be a yaml file or directory.")

        schemas = []
        for f in all_files:
            with open(f, "r") as inp_file:
                schema = yaml.safe_load(inp_file)
                if "type" in schema:
                    if schema["type"].lower() == "extension":
                        # Need to use environments extension manager
                        current_env = self.client.environments.get(self.parent)
                        schema = current_env.extensions.push_extension(
                            os.path.dirname(f), schema=schema
                        )
                else:
                    raise Exception(f"Mandatory field 'type' not found in {f}.")
                schemas.append(yaml.dump(schema))

        if yaml_text is not None:
            schemas.append(yaml_text)

        req = management_pb2.CreateChangeRequest(
            parent=self.parent,
            schemas=schemas,
            plan_only=plan_only,
            purge=purge,
            profile_none=profile_none,
            update_all=update_all,
            train_all=train_all,
            train_none=train_none,
            message=message,
            trigger=trigger,
            resource=resource,
        )
        resp = self.client._management.CreateChange(req)
        return Change.from_proto(resp, client=self.client)

    def rerun(self, id: str):
        req = management_pb2.ReRunRequest(name=self.name(id))
        resp = self.client._management.ReRun(req)
        return Change.from_proto(resp, client=self.client)

    def get(self, id: str):
        req = management_pb2.GetChangeRequest(name=self.name(id))
        resp = self.client._management.GetChange(req)
        return Change.from_proto(resp, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        filters: List[str] = None,
        all_projects: bool = False,
    ) -> List[Change]:
        """List all changes in the project.

        Arguments:
            page_size: Number of items to return.

        Returns:
            A list of changes.
        """
        req = management_pb2.ListChangesRequest(
            parent=self.parent,
            page_size=page_size,
            filters=filters,
            all_projects=all_projects,
        )
        resp = self.client._management.ListChanges(req)
        return [Change.from_proto(x, client=self.client) for x in resp.plans]

    def list_all(self) -> Iterator[Change]:
        """List all changes.

        Pages through all changes using an iterator.

        Returns:
            A iterator of all changes.
        """

        def next_page(next_page_token):
            req = management_pb2.ListChangeRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListChange(req)
            return (
                [Change.from_proto(x, client=self.client) for x in resp.plans],
                resp.next_page_token,
            )

        return Pager(next_page)

    def cancel(self, id: str) -> None:
        """Cancel an change workflow.

        Arugments:
            id: Name or id of change workflow

        """
        req = management_pb2.CancelChangeRequest(name=self.name(id))
        self.client._management.CancelChange(req)


class Change(Resource, types.Change):
    """Change resource."""

    name_pattern: str = "projects/{project}/changes/{change}"
    manager: ChangeManager
    events: EventManager
    """Event manager."""

    def _init(self):
        self.manager = ChangeManager(parent=self.parent, client=self.client)
        self.events = EventManager(parent=self.name, client=self.client)

    def cancel(self):
        """Cancel change."""
        self.manager.cancel(self.id)

    def rerun(self):
        """Rerun changes."""
        return self.manager.rerun(self.id)

    def wait(self, echo: bool = True, timeout: int = None) -> Change:
        """Wait for Change operation to complete.

        Arguments:
            echo: Display progress.
            timeout: Timeout in seconds.
        Returns:
            Updated changes.
        """
        # if it has already succeeded, return immediately
        if self.state in ["SUCCEEDED", "CANCELLED", "FAILED"]:
            return self

        try:
            changes = wait(
                getter=lambda: self.manager.get(self.id),
                echo=echo,
                timeout=timeout,
            )
            return changes
        except FailedError as e:
            # Get latest change and create a dict of error message for first failure
            change = self.manager.get(self.id)
            plan_step_state_values = [s.state.value for s in change.plan]
            failures = {}
            for i in range(len(plan_step_state_values)):
                if plan_step_state_values[i] == "FAILED":
                    failure_step_type = change.plan[i].type.name
                    failure_error_messages = change.plan[i].error_messages
                    failures[failure_step_type] = failure_error_messages
            e.details.update({"failures": failures})
            raise FailedError(e.message, details=e.details)



from __future__ import annotations
import yaml
import toml
import os
import shutil
from pathlib import Path
from typing import Optional, List, Tuple
import subprocess
from glob import glob
import requests
import hashlib
import sys
import tarfile

from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.events import EventManager
from continual.python.sdk.extensions import templates


TEMPLATES_DIR_PATH = os.path.dirname(templates.__file__)
DEFAULT_TEMPLATE_CLASS_BY_TYPE = {
    "algorithm": "MyCustomAlgorithm",
    "model": "MyModelPipeline",
}


class ExtensionManager(Manager):
    """Manages promotion resources."""

    name_pattern: str = "projects/{project}/extensions/{extension}"

    def get(self, id: str) -> Extension:
        """
        Get extension

        Parameters
        ----------
        id: str
            Extension id

        Returns
        -------
        An extension
        """
        req = management_pb2.GetExtensionRequest(name=self.name(id))
        resp = self.client._management.GetExtension(req)
        return Extension.from_proto(resp, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        all_projects: bool = False,
    ) -> List[Extension]:
        """
        List extensions

        Parameters
        ---------
        page_size: int
            Number of items to return
        all_projects: bool
            Whether to show all projects or not
        """

        req = management_pb2.ListExtensionsRequest(
            parent=self.parent,
            page_size=page_size,
            all_projects=all_projects,
        )
        resp = self.client._management.ListExtensions(req)
        return [Extension.from_proto(x, client=self.client) for x in resp.extensions]

    def delete(self, id: str):
        """
        Delete an extension

        Parameters
        ----------
        id : str
            Name of the extension to delete
        """

        req = management_pb2.DeleteExtensionRequest(name=self.name(id))
        self.client._management.DeleteExtension(req)

    def export(self, id: str, path: Path, extract: bool) -> str:
        """
        Get the source code tarball
        for this extension and possibly untar
        it

        Parameters
        ----------
        id : str
            Name of the extension
        path: Path
            Path to which to download the source tarball
        extract: bool
            Whether or not to extract the contents of the
            tarball

        Returns
        -------
        str
            Path to exported file/directory
        """

        req = management_pb2.ExportExtensionRequest(name=self.name(id))
        signed_url = self.client._management.ExportExtension(req)

        res = requests.get(signed_url.signed_url, stream=True)
        new_file_path = os.path.join(path, f"{id}.tar.gz")
        with open(new_file_path, "wb") as f:
            f.write(res.content)
        if extract:
            with tarfile.open(name=new_file_path, mode="r") as f:
                old_file_path = new_file_path
                new_file_path = os.path.join(path, id)
                f.extractall(new_file_path)
            os.remove(old_file_path)
        return new_file_path

    def install(self, id: str):
        """
        Pip install extension python package using straight from signed URL

        Parameters
        ----------
        id : str
            Name of the extension

        """

        req = management_pb2.ExportExtensionRequest(name=self.name(id))
        signed_url = self.client._management.ExportExtension(req)

        # Uninstall older version of extension if any
        try:
            subprocess.run(
                [
                    sys.executable,
                    "-m",
                    "pip",
                    "uninstall",
                    "-y",
                    id.lower(),
                ],
                check=True,
                capture_output=True,
            )
        except subprocess.CalledProcessError as e:
            raise Exception(
                f"Could not uninstall older version of extension {id} : {str(e.output, encoding='utf-8')}, {str(e.stderr, encoding='utf-8')}, {str(e.stdout, encoding='utf-8')}"
            )

        # Now install it
        try:
            subprocess.run(
                [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    signed_url.signed_url,
                ],
                check=True,
                capture_output=True,
            )
        except subprocess.CalledProcessError as e:
            raise Exception(
                f"Could not install extension {id} : {str(e.output, encoding='utf-8')}, {str(e.stderr, encoding='utf-8')}, {str(e.stdout, encoding='utf-8')}"
            )

    def init(
        self,
        root: str,
        id: str,
        extension_type: str,
        class_name: str,
        source_path: str,
        dependencies: List[str],
    ):
        """
        Creates a template pip-managed python project for a Continual Extension
        with given parameters

        Parameters
        ----------
        root: str
            Absolute path to directory in which to create the project
        id: str
            ID of the extension, used in other resource YAMLs to identify extension
        extension_type: str
            Type of the extension. Must be one of the supported types
        class_name: str
            The name of the class that will contain the model implementation
        source_path: str
            The optional path to the source file implementation to copy into the directory,
            if not specified a template implementation will be copied into the extension directory
        dependencies: List[str]
            Names of pip package dependencies (with versions attached) to be passed to poetry add
        """

        ext_dir = os.path.join(root, id)

        # Run poetry new
        try:
            # Create the directory
            subprocess.run(
                ["poetry", "new", id, "--name", id],
                check=True,
                capture_output=True,
                cwd=root,
            )
        except subprocess.CalledProcessError as e:
            raise Exception(
                f"Could not create new package {id} : {str(e.output, encoding='utf-8')}"
            )

        # Add package dependencies to project file if any
        if dependencies:
            try:
                subprocess.run(
                    ["poetry", "add", "--lock"] + dependencies,
                    check=True,
                    capture_output=True,
                    cwd=ext_dir,
                )
            except subprocess.CalledProcessError as e:
                raise Exception(
                    f"Could not add package dependencies for {id} : {str(e.output, encoding='utf-8')}"
                )

        self._copy_template(root, id, extension_type, source_path, class_name)
        self._add_extension_schema(root, id, class_name, extension_type)

    def _copy_template(
        self,
        root: str,
        extension_id: str,
        extension_type: str,
        source_path: str,
        class_name: str,
    ):
        """
        Create a template in the extension  python package to faciliate dev
        """

        ext_dir = os.path.join(root, extension_id)

        # The name needs to be lowercase here because of the way poetry creates directories
        dest_file_path = os.path.join(
            ext_dir, extension_id.lower(), f"{extension_type}.py"
        )
        if not source_path:
            src_file_path = os.path.join(
                TEMPLATES_DIR_PATH, f"template_custom_{extension_type}.py"
            )
            with open(src_file_path, "r") as fin, open(dest_file_path, "w+") as fout:
                template_contents = fin.read()
                if class_name:
                    template_contents = template_contents.replace(
                        DEFAULT_TEMPLATE_CLASS_BY_TYPE[extension_type],
                        class_name,
                    )
                fout.write(template_contents)
                fout.flush()
                os.fsync(fout)
        else:
            # Copy over existing implementation file
            shutil.copy(source_path, dest_file_path)

    def _add_extension_schema(
        self, root: str, extension_id: str, class_name: str, extension_type: str
    ):
        """
        Create an extension yaml for a custom extension at path

        Parameters
        ----------
        path: str
            The root directory of the extension source tree
        extension_id: str
            The name/label for the extension
        class_name: str
            The name of the extension implementation
        """

        schema = {
            "type": "extension",
            "extension_type": extension_type,
            "extension_id": extension_id,
            "module_name": ".".join([extension_id.lower(), extension_type]),
            "class_name": class_name
            if class_name
            else DEFAULT_TEMPLATE_CLASS_BY_TYPE[extension_type],
        }

        with open(os.path.join(root, extension_id, "extension.yaml"), "w") as f:
            yaml.dump(schema, f)

    def push_extension(self, root: str, schema: dict) -> dict:
        """
        Prepare package and upload into continual

        Parameters
        ----------
        root: str
            the path to the package root directory
        schema: dict
            yaml config/schema for the extension rooted at
            root.e
        """

        new_schema = schema.copy()

        # Get the full resource name
        new_schema["name"] = self.name(schema["extension_id"])

        # Package if necessary, and attach package md5 and resource_url to schema
        package_path = self._package_extension(root=root, schema=new_schema)
        new_schema["package_hash"] = self._create_package_hash(
            package_path=package_path
        )
        signed_url, new_schema["package_url"] = self._get_extension_bucket_info(
            extension_id=schema["extension_id"],
            package_filename=Path(package_path).name,
        )

        # Upload file
        resp = requests.put(
            url=signed_url,
            data=open(package_path, "rb").read(),
            headers={"Content-Type": "application/gzip"},
        )
        if resp.text != "":
            raise Exception(
                f"Could not upload package {package_path} to Continual: {resp.text}"
            )
        return new_schema

    def _package_extension(self, root: str, schema: dict) -> str:
        """
        Packages a particular extension given the yaml

        Parameters
        ----------
        root: str
            the path to the package root directory
        schema: dict
            yaml config/schema for the extension rooted at
            root.

        Returns
        -------
        str
            absolute path to distributable package file

        Raises
        ------
        Exception
            If packaging process failed
        """

        # If the package-path is not specified, run the default package manager
        if not "package_path" in schema:
            # Add YAMLs to package data
            try:
                pyproject_filepath = os.path.join(root, "pyproject.toml")
                pyproject_file = toml.load(pyproject_filepath)
                pyproject_file["tool"]["poetry"]["include"] = ["*.yaml"]
                with open(pyproject_filepath, "w") as f:
                    toml.dump(pyproject_file, f)
            except Exception as e:
                raise Exception(
                    f"pyproject file could not be edited to include package data: {e}"
                )
            # poetry build
            try:
                subprocess.run(
                    ["poetry", "build"], check=True, capture_output=True, cwd=root
                )
            except subprocess.CalledProcessError as e:
                raise Exception(
                    f"Could not build package '{schema['name']}' for push: {str(e.output, encoding='utf-8')}"
                )

            dist_packages = glob(os.path.join(root, "dist", "*.tar.gz"))
            if len(dist_packages) != 1:
                raise Exception(
                    f"Package build failed. Remove old .tar.gz files from {os.path.join(root, 'dist')}"
                )
            else:
                package_path = dist_packages[0]
        else:
            package_path = schema["package_path"]

        return package_path

    def _create_package_hash(self, package_path: str) -> str:
        """
        Creates an md5 hash of contents of distributable
        package file

        Parameters
        ----------
        package_path: str
            absolute path to distributable package file

        Returns
        -------
        str
            md5 hash as a hex string of the file at package_path
        """

        hash_md5 = hashlib.md5()
        with open(package_path, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

    def _get_extension_bucket_info(
        self, extension_id: str, package_filename: str
    ) -> Tuple[str, str]:
        """
        Get a bucket url and push id for the extension package

        Parameters
        ----------
        extension_id: str
            name of the extension
        package_filename: str
            Filename to use in bucket for packaged extension

        Returns
        -------
        (str,str)
            (signed_url,bucket_url) for this package
        """
        req = management_pb2.GetExtensionBucketPathRequest(
            parent=self.parent, extension_id=extension_id, file_name=package_filename
        )
        resp = self.client._management.GetExtensionBucketPath(req)
        return resp.signed_url, resp.bucket_path


class Extension(Resource, types.Extension):
    """Extension resource."""

    name_pattern: str = "projects/{project}/extensions/{extension}"
    manager: ExtensionManager
    """ExtensionManager"""

    events: EventManager
    """Event manager."""

    def _init(self):
        self.manager = ExtensionManager(parent=self.parent, client=self.client)
        self.events = EventManager(parent=self.name, client=self.client)



import sys
import logging
import os
from continual.python.blob.store import BlobStore

LOGGING_FMT = "%(asctime)s %(levelname)-4s" "[%(filename)s:%(lineno)d]  %(message)s"
LOGGING_DATEFMT = "%Y-%m-%d %H:%M:%S"


class OutputSplicer:
    """
    We set stdout and stderr to one of
    these object so that writes are duplicated to a log file as well
    as to stdout and stderr
    """

    def __init__(self, log_path: str, buffered: bool = False):
        """
        Parameters
        ----------
        log_path: str
            path to the log
        buffered: bool
            Whether to buffer output
        """
        self._log_file = open(log_path, "a+")
        self._buffered = buffered

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Exit 'with' context
        """
        self.close()

    def close(self):
        """
        Close file like object
        """
        self._log_file.close()

    def write(self, message):
        """
        Write to
        """
        self._log_file.write(message)
        if not self._buffered:
            self._log_file.flush()

    def flush(self):
        """
        Flush buffered output
        """
        self._log_file.flush()


class StandardOutputSplicer(OutputSplicer):
    """
    Context manager for regions where we want stdout to also go to a log file
    """

    def __enter__(self):
        """
        Enter 'with' context
        """
        self._stdout = sys.stdout
        sys.stdout = self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Exit 'with' context
        """
        super().__exit__(exc_type, exc_value, traceback)
        sys.stdout = self._stdout

    def write(self, message):
        """
        Write to output
        """
        super().write(message)
        self._stdout.write(message)

    def flush(self):
        """
        Flush buffered output
        """
        super().flush()
        self._stdout.flush()


class StandardErrorSplicer(OutputSplicer):
    """
    Context manager for regions where we want stderr to also go to a log file
    """

    def __enter__(self):
        """
        Enter 'with' context
        """
        self._stderr = sys.stderr
        sys.stderr = self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Exit 'with' context
        """
        super().__exit__(exc_type, exc_value, traceback)
        sys.stderr = self._stderr

    def write(self, message):
        """
        Write to output
        """
        super().write(message)
        self._stderr.write(message)

    def flush(self):
        """
        Flush buffered output
        """
        super().flush()
        self._stderr.flush()


def configure_logging(master_log_path: str):
    """
    Set logging module parameters

    Parameters
    ----------
    master_log_path: str
        The master log where the output of all loggers will go
    """
    logging.basicConfig(
        level=logging.INFO,
        format=LOGGING_FMT,
        datefmt=LOGGING_DATEFMT,
        filename=master_log_path,
        force=True,
    )

    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setFormatter(
        logging.Formatter(fmt=LOGGING_FMT, datefmt=LOGGING_DATEFMT)
    )
    logging.root.addHandler(stdout_handler)

    logging.info(
        f"Configure training service logger to write to {os.path.abspath(master_log_path)}"
    )


def get_logger(name: str, log_file_path: str, log_level: str = "INFO"):
    """
    Return a logger with the give name that logs to a filepath

    Parameters
    ----------
    name : str
        Name of the logger
    log_file_path: str
        The path to the file that the logger will log to
    log_level: str
        The log level for the returned logger
    """

    # Create handler
    formatter = logging.Formatter(fmt=LOGGING_FMT, datefmt=LOGGING_DATEFMT)
    fh = logging.FileHandler(filename=log_file_path)
    fh.setFormatter(formatter)

    # Add handler
    logger = logging.getLogger(name)
    logger.addHandler(fh)
    logger.setLevel(log_level)

    return logger


def upload_logs(file_path: str, bucket: str, bucket_relative_path: str) -> None:
    store = BlobStore(bucket)
    store.upload_file(file_path, prefix=bucket_relative_path)
    return
